/*
 * Sercomm suzhou
 * Add a bridge to bridge traffic betwenn Ethernet and ATM
 * 
 * Solomon shuang@sdc.sercomm.com
 * 2006/07/07
 * No distribution allowed unless under Sercomm approval
 * 
 * change list:
 * $Log: lwb_main.c,v $
 * Revision 1.7  2008-01-15 06:36:23  laoyou_liu
 * do not check source
 *
 * Revision 1.6  2007-07-14 10:42:32  laoyou_liu
 * don't check ip, just bridge. fix the tunnel subnet bug
 *
 * Revision 1.5  2007/03/02 05:14:35  laoyou_liu
 * pppoe exhaust resource
 *
 * Revision 1.4  2006/09/19 11:23:42  shin
 * fix wan dod fail
 *
 * Revision 1.3  2006/07/14 08:58:36  shin
 * remove debug flag
 *
 * Revision 1.2  2006/07/14 04:16:29  shin
 * 1 change ip_forward to ip_send, so no ttl decrease
 * 2 change wan_dev to lwb_wan_dev, lan_dev to lwb_lan_dev
 *
 * Revision 1.1  2006/07/11 13:13:59  shin
 * add lan wan bridge in kernel
 *
*/

#include <linux/kernel.h>
//#include <linux/module.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/brlock.h>
#include <linux/etherdevice.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/if_ether.h>
#include <linux/if_arp.h>
#include <linux/inetdevice.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/route.h>
#include <net/ip.h>

//#define LWB_DEBUG

#ifdef LWB_DEBUG
#define DK(fmt,arg...) printk(fmt,##arg)
#else
#define DK(fmt,arg...)
#endif
struct dev_count
{
    unsigned long send;
    unsigned long drop;
};

enum lwb_error
{
    NO_ERROR,
    NO_LANDEV,
    NO_WANDEV,
    CRT_PROC,
    CMD_ERR,
    SKB_ARPREQ_NULL,
    SKB_SIP_NULL,
    SKB_DIP_NULL,
    DEV_IP_NULL
};

struct lwb_stat
{
    struct dev_count lan;
    struct dev_count wan;
    unsigned long start;
    unsigned long stop;
    enum lwb_error err;
};

extern struct sk_buff *(*lwb_frame_hook)(struct sk_buff *skb);
extern struct net_device *lwb_wan_dev,*lwb_lan_dev;

static char _wan_type;
static char _wan_if[10];
static struct lwb_stat _lwb_stat;
static char _mac_addr[6];
static char _lwb_dod=0;

static void lwb_forward2_lanip(struct sk_buff *skb);
static void lwb_forward2_wanip(struct sk_buff *skb);
static void lwb_forward2_wanother(struct sk_buff *skb);
static void lwb_forward2_wanppp(struct sk_buff *skb);
/* static void lwb_forward2_wanipoa(struct sk_buff *skb); */
static struct sk_buff * lwb_drop_lan(struct sk_buff *skb);
static struct sk_buff * lwb_drop_wan(struct sk_buff *skb);
static struct sk_buff * build_arp_rsp(struct sk_buff *skb);
static struct sk_buff * lwb_handle_frame (struct sk_buff *skb);

static inline int SKB_IP(struct sk_buff *skb)
{
    DK("SKB_IP proto is %2.2x\n",ntohs(skb->protocol));
    return (ntohs(skb->protocol)==ETH_P_IP);
}

static inline int SKB_ARPREQ(struct sk_buff *skb)
{
    DK("SKB_ARPREQ (%2.2x)(%2.2x)\n",ntohs(skb->protocol),skb->nh.arph?ntohs(skb->nh.arph->ar_op):(short)-1);
    if (!skb->nh.arph)
    {
        DK("No ARP OP\n");
        _lwb_stat.err = SKB_ARPREQ_NULL;
        return 0;
    }
    return ((ntohs(skb->protocol)==ETH_P_ARP)&&(ntohs(skb->nh.arph->ar_op) ==ARPOP_REQUEST));
}

static inline unsigned long SKB_SIP(struct sk_buff *skb)
{
    DK("SKB_SIP %4.4x\n",ntohl(skb->nh.iph->saddr));
    if (!skb->nh.iph)
    {
        DK("No SKB_SIP\n");
        _lwb_stat.err = SKB_SIP_NULL;
        return 0;
    }
    return ntohl(skb->nh.iph->saddr);
}

static inline unsigned long SKB_DIP(struct sk_buff *skb)
{
    DK("SKB_DIP %4.4x\n",ntohl(skb->nh.iph->daddr));
    if (!skb->nh.iph)
    {
        DK("No SKB_DIP\n");
        _lwb_stat.err = SKB_DIP_NULL;
        return 0;
    }
    return (unsigned long)ntohl(skb->nh.iph->daddr);
}

static inline unsigned long DEV_IP(struct net_device *dev)
{
    DK("DEV_IP dev(%p) ip(%x)\n",dev,ntohl(dev?dev->ip_ptr?((struct in_device *)(dev->ip_ptr))->ifa_list?(((struct in_device *)(dev->ip_ptr))->ifa_list->ifa_local):0:0:0));
    if (!dev||!dev->ip_ptr||!((struct in_device *)(dev->ip_ptr))->ifa_list)
    {
        DK("No DEV_IP\n");
        _lwb_stat.err = DEV_IP_NULL;
        return 0;
    }
    return ntohl(((struct in_device *)(dev->ip_ptr))->ifa_list->ifa_local);
}

#define release_if(n) do { if (n) dev_put(n); n=NULL;} while (0)

static void __lwb_hook_up(void)
{
    lwb_lan_dev = dev_get_by_name("br0");
    if ( !lwb_lan_dev )
    {
        DK("lwb: fail to up, no lan device\n");
        _lwb_stat.err = NO_LANDEV;
        return;
    }
    lwb_wan_dev = dev_get_by_name(_wan_if);
    if ( !lwb_wan_dev )
    {
        DK("lwb: fail to up, no wan device\n");
        _lwb_stat.err = NO_WANDEV;
        release_if(lwb_lan_dev);
        return;
    }
    lwb_frame_hook = lwb_handle_frame;
    _lwb_stat.start++;
}

static void __lwb_hook_down(void)
{
    release_if(lwb_lan_dev);
    release_if(lwb_wan_dev);
    lwb_frame_hook = NULL;
    _lwb_stat.stop++;
}

void  lwb_hook_up(void)
{
    br_write_lock_bh(BR_NETPROTO_LOCK);
    __lwb_hook_up();
    br_write_unlock_bh(BR_NETPROTO_LOCK);
}

void lwb_hook_down(void)
{
    br_write_lock_bh(BR_NETPROTO_LOCK);
    __lwb_hook_down();
    br_write_unlock_bh(BR_NETPROTO_LOCK);
}

static int lwb_proc_write (struct file*file, const char*buffer,unsigned long count,void* data)
{
    int ret;
    static char _cmd[20];
    static char _temp_if[20];
    static char _temp_type[20];
    static int _start;
    
    if (count > 20)
        return -EFAULT; 

    /* Initialize the buffer before using it. */
    memset ((void *)_cmd, 0, sizeof(_cmd));

    _start = -1;
    
    /* Copy from user space. */	
    if (copy_from_user (_cmd, buffer, count))
        goto error;
    DK("lwb:input <%s>\n",_cmd);
    ret = sscanf(_cmd,"wan=%s type=%s",_temp_if,_temp_type);
    /* parse config command sucessfully */
    if ( ret == 2 )
    {
        DK("lwb:wan=%s,type=%c\n",_temp_if,_temp_type[0]);
        if (!_temp_if[0]||(strlen(_temp_if)>4))
        {
            DK("lwb: wrong wan device\n");
            goto error;
        }
        strcpy(_wan_if,_temp_if);
        if ( (_temp_type[0]!='p')
             &&(_temp_type[0]!='i')
             &&(_temp_type[0]!='e'))
        {
            DK("lwb: unknown type %c, should be [p|i|e], as PPP|IP|Ethernet\n",_temp_type[0]);
            goto error;
        }
        _wan_type=_temp_type[0];
        return count;
    }

    ret = sscanf(_cmd,"start=%d",&_start);
    /* parse start/stop command sucessfully */
    if ( ret == 1 )
    {
        DK("lwb:start=%d\n",_start);
        if ( _start == 1 )
        {
            _lwb_dod = 0;
            lwb_hook_up();
        }
        else if ( _start == 0 )
        {
            _lwb_dod = 0;
            lwb_hook_down();
        }
        else if ( _start == 2 )
        {
            _lwb_dod =1;
            lwb_hook_up();
        }
        else
            goto error;

        return count;
    }

    DK("lwb:fail to parse command\n");
error:
    _lwb_stat.err = CMD_ERR;
    return -EFAULT;
}

int __init lwb_init(void)
{
    struct proc_dir_entry*  ptr_dir_entry;
    printk ("LWB: LAN-WAN bridge version 1.0..");

    ptr_dir_entry = create_proc_entry("net/lwb", 0644, NULL);
    if (ptr_dir_entry == NULL)
    {
        printk ("ERROR: Unable to create the proc entry for LAN-WAN bridge creation.\n");
        _lwb_stat.err = CRT_PROC;
        return -1;
    }
    ptr_dir_entry->data         = NULL;
    ptr_dir_entry->write_proc   = lwb_proc_write;
    ptr_dir_entry->owner        = NULL;

    lwb_frame_hook = NULL;
    _wan_type='e';
    memset(_mac_addr,0xff,ETH_ALEN);    

    printk ("Done\n");
    return 0;
}

static void __exit lwb_deinit(void)
{
    lwb_hook_down();
    remove_proc_entry("net/lwb", NULL);
    return;
}

/* #define NOT_IP_ARP_RESPONSE_TO_LOCAL(skb) (((ntohl(skb->nh.iph->daddr))!=0xc0a80001)\ */
/*                                                 &&(skb->nh.arph->ar_op!=ARPOP_REPLY))                          */
/* filter out skbs we shouldn't take care of
 return:
 1 : filter out
 0 : pass
*/
static int lwb_filter(struct sk_buff *skb)
{
    if ( skb->dev == lwb_lan_dev )
    {
        /* it is from br0 and it is for br0*/
        if ( SKB_DIP(skb)==DEV_IP(lwb_lan_dev))
            return 1;
        /* it is ARP reply, we need to give it to br0
         since sometimes br0 will send ARP request*/
        if ( (skb->protocol == htons(ETH_P_ARP))&&
             (skb->nh.arph->ar_op==ARPOP_REPLY))
            return 1;
    }
    else
    {
        /* if it is a DHCP reply, we gonna pass it,
           nas0's dhcpc need it*/
        unsigned char *tmp;
        tmp = (unsigned char *)skb->nh.iph;
        tmp += (skb->nh.iph->ihl)<<2;
        if ( (skb->nh.iph->protocol==17)&&
             (tmp[0]==0)&&(tmp[1]==0x43)&&
             (tmp[2]==0)&&(tmp[3]=0x44))
            return 1;
    }
    return 0;
}

static struct sk_buff * lwb_handle_frame (struct sk_buff *skb)
{
    DK("lwb_handle_frame: input skb <%p> (%c)mode\n",skb,_wan_type);
    if (!skb) return NULL;
    if ( lwb_filter(skb) ) return skb;
    
    if ( skb->dev == lwb_lan_dev )
    {
        /* received from lan, we gonna forward it to wan or drop is */
        if (_wan_type=='e')
        {
            /* dhcpc or fixed IP */
            /* if (SKB_SIP(skb)==DEV_IP(lwb_wan_dev)) */
            lwb_forward2_wanip(skb);
            return NULL;
        }
        else
        {
            /* in case of pppoe/pppoa/ipoa */
            if (SKB_ARPREQ(skb))
            {
                /* response ARP request */
                skb = build_arp_rsp(skb);
                lwb_forward2_lanip(skb);
                return NULL;
            }
            else if ( SKB_IP(skb)&&
                        (_lwb_dod||(SKB_SIP(skb)==DEV_IP(lwb_wan_dev))))
            {
                /* forward it */
               /* if (_wan_type=='p'||_wan_type=='i'){
                    lwb_forward2_wanother(skb);
                }*/
                if (_wan_type=='p'){
                    lwb_forward2_wanppp(skb);
                }
                else if(_wan_type=='i'){
                    lwb_forward2_wanother(skb);
                }
                else /* safe code */
                    return lwb_drop_lan(skb);
                return NULL;
            }
            return lwb_drop_lan(skb);
        }
    }
    else
    {
        /* received from wan, we gonna forward it to lan or drop it*/
        if (_wan_type=='e')
        {
            /* dhcpc or fixed IP */
            lwb_forward2_lanip(skb);
            return NULL;
        }
        else
        {
            /* in case of pppoe/pppoa/ipoa */
            if ( SKB_IP(skb)/*&&(SKB_DIP(skb)==DEV_IP(lwb_wan_dev))*/)/*tunnel subnet doesn't work*/
            {
                if (skb_headroom(skb) < ETH_HLEN)
                {
                    struct sk_buff *ns;
                    DK("add 14 bytes for ether head\n");
                    ns = alloc_skb(skb->len + ETH_HLEN, GFP_ATOMIC);
                    if (ns == 0)
                    {
                        kfree_skb(skb);
                        return NULL;
                    }
                    skb_reserve(ns, ETH_HLEN);
                    memcpy(skb_put(ns, skb->len), skb->data, skb->len);
                    ns->nh.raw = ns->data;
                    kfree_skb(skb);
                    skb = ns;
                }
                skb->mac.raw = skb->nh.raw-ETH_HLEN;
                memcpy(skb->mac.ethernet->h_dest,_mac_addr,ETH_ALEN);
                memcpy(skb->mac.ethernet->h_source,lwb_lan_dev->dev_addr,ETH_ALEN);
                skb->mac.ethernet->h_proto=htons(ETH_P_IP);
                lwb_forward2_lanip(skb);
                return NULL;
            }
            else
                return lwb_drop_wan(skb);
        }
    }
}

static void lwb_forward2_lanip(struct sk_buff *skb)
{
    DK("lwb_forward2_lanip: enter\n");
    skb_push(skb,ETH_HLEN);
    skb->dev = lwb_lan_dev;
    dev_queue_xmit(skb);
    _lwb_stat.lan.send++;
}

static void lwb_forward2_wanip(struct sk_buff *skb)
{
    DK("lwb_forward2_wanip: enter\n");
    skb_push(skb,ETH_HLEN);
    skb->dev = lwb_wan_dev;
    dev_queue_xmit(skb);
    _lwb_stat.wan.send++;
}
#if 1
static void lwb_forward2_wanppp(struct sk_buff *skb)
{
    DK("lwb_forward2_wanppp: enter\n");
    skb->dev = lwb_wan_dev;
    dev_queue_xmit(skb);
    _lwb_stat.wan.send++;
}
#endif

static void lwb_forward2_wanother(struct sk_buff *skb)
{
    struct rtable *rt;
    int error;

    DK("lwb_forward2_wanipoa: enter\n");
    skb->dev = lwb_wan_dev;
    error = ip_route_output(&rt,skb->nh.iph->daddr,0,0,0);
    if (error) goto drop;
    skb->dst = rt;
    //ip_forward(skb);
    ip_send(skb);
    ip_rt_put(rt);
/*     neigh = __neigh_lookup(&clip_tbl,&skb->nh.iph->daddr,rt->u.dst.dev,1); */
/*     ip_rt_put(rt); */

/*     if (!neigh) goto drop; */
/*     skb->dst = rt; */
/*     skb->dst->neighbour = neigh; */
/*     neigh_release(neigh); */
/*     dev_queue_xmit(skb); */
    _lwb_stat.wan.send++;
    return;
drop:
    kfree_skb(skb);
    return;
}

static struct sk_buff *lwb_drop_lan(struct sk_buff *skb)
{
    DK("lwb_drop_lan: drop\n");
    //kfree_skb(skb);
    _lwb_stat.lan.drop++;
    return skb;
}

static struct sk_buff * lwb_drop_wan(struct sk_buff *skb)
{
    DK("lwb_drop_wan: drop\n");
    //kfree_skb(skb);
    _lwb_stat.wan.drop++;
    return skb;
}

static struct sk_buff *build_arp_rsp(struct sk_buff *skb)
{
    char _tmpip[4];
    struct _arp_plh
    {
	unsigned char		ar_sha[ETH_ALEN];	/* sender hardware address	*/
	unsigned char		ar_sip[4];		/* sender IP address		*/
	unsigned char		ar_tha[ETH_ALEN];	/* target hardware address	*/
	unsigned char		ar_tip[4];		/* target IP address		*/
    }* arp_pl;
    arp_pl=(struct _arp_plh *)(skb->nh.arph+1);
    
    skb->nh.arph->ar_op = htons(ARPOP_REPLY);
    memcpy(skb->mac.ethernet->h_dest,skb->mac.ethernet->h_source,ETH_ALEN);
    memcpy(skb->mac.ethernet->h_source,lwb_lan_dev->dev_addr,ETH_ALEN);

    memcpy(arp_pl->ar_tha,arp_pl->ar_sha,ETH_ALEN);
    memcpy(arp_pl->ar_sha,lwb_lan_dev->dev_addr,ETH_ALEN);
    memcpy(_tmpip,arp_pl->ar_tip,4);
    memcpy(arp_pl->ar_tip,arp_pl->ar_sip,4);
    memcpy(arp_pl->ar_sip,_tmpip,4);

    /*record dest MAC address */
    memcpy(_mac_addr,skb->mac.ethernet->h_dest,ETH_ALEN);    
    return skb;
}

/* EXPORT_NO_SYMBOLS; */
module_init(lwb_init)
module_exit(lwb_deinit)

